SystemOrganization addCategory: #'Neo-JSON-Core'!

Object subclass: #NeoJSONReader
	instanceVariableNames: 'readStream stringStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-JSON-Core'!

!NeoJSONReader commentStamp: 'SvenVanCaekenberghe 6/2/2012 13:49' prior: 0!
I am NeoJSONReader.!

!NeoJSONReader class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 6/2/2012 14:03'!
on: readStream
	^ self new
		on: readStream;
		yourself! !

!NeoJSONReader methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 6/2/2012 13:50'!
close
	readStream ifNotNil: [
		readStream close.
		readStream := nil ]! !

!NeoJSONReader methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/2/2012 21:26'!
consumeWhitespace
	"Strip whitespaces from the input stream."

	[ readStream atEnd not and: [ readStream peek isSeparator ] ]
		whileTrue: [ 
			readStream next ]! !

!NeoJSONReader methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/2/2012 14:01'!
expectChar: character
	"Expect character and consume input and optional whitespace at the end,
	 throw an error otherwise."

	(self matchChar: character)
		ifFalse: [ self error: character asString, ' expected' ]! !

!NeoJSONReader methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 6/2/2012 16:18'!
listClass
	^ Array! !

!NeoJSONReader methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 6/2/2012 16:18'!
mapClass
	^ Dictionary! !

!NeoJSONReader methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/2/2012 14:16'!
match: string do: block
	"Try to read and consume string and execute block if successful.
	Else do nothing (but do not back up)"
	 
	(string allSatisfy: [ :each | self matchChar: each ] )
		ifTrue: [ block value ]
! !

!NeoJSONReader methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/2/2012 14:01'!
matchChar: character
	"Tries to match character, consume input and 
	answer true if successful and consumes whitespace at the end."
	
	^ readStream peek = character
		ifTrue: [
			readStream next.
			self consumeWhitespace.
			true ]
		ifFalse: [
			false ]! !

!NeoJSONReader methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 6/2/2012 13:52'!
next
	^ self 
		consumeWhitespace; 
		parseValue! !

!NeoJSONReader methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 6/2/2012 13:50'!
on: aReadStream
	readStream := aReadStream! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 21:45'!
parseCharacter
	| char |
	(char := readStream next) = $\ 
		ifFalse: [ ^ char ].
	(#( $" $/ $\) includes: (char := readStream next))
		ifTrue: [ ^ char ].
	('bfnrt' includes: char)
		ifTrue: [ ^ self unescapeChar: char ].
	char = $u
		ifTrue: [ ^ self parseCharacterHex ].
	self error: 'invalid escape character \' , (String with: char)! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 14:20'!
parseCharacterHex
	| value |
	value := self parseCharacterHexDigit.
	3 timesRepeat: [ 
		value := (value << 4) + self parseCharacterHexDigit ].
	^ Character codePoint: value! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 14:21'!
parseCharacterHexDigit
	| digit |
	readStream atEnd 
		ifFalse: [ 
			digit := readStream next asInteger.
			(digit between: "$0" 48 and: "$9" 57)
				ifTrue: [ ^ digit - 48 ].
			(digit between: "$A" 65 and: "$F" 70)
				ifTrue: [ ^ digit - 55 ].
			(digit between: "$a" 97 and: "$f" 102)
				ifTrue: [ ^ digit - 87 ] ].
	self error: 'hex-digit expected'! !

!NeoJSONReader methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/2/2012 14:16'!
parseConstantDo: block
	"Parse and consume either true|false|nil|null and execute block 
	or else do nothing (but do not back up).
	Hand written implementation to avoid the use of #position:"
	
	(readStream peek = $t)
		ifTrue: [
			^ self match: 'true' do: [ block value: true ] ].
	(readStream peek = $f)
		ifTrue: [
			^ self match: 'false' do: [ block value: false ] ].
	(readStream peek = $n)
		ifTrue: [
			^ self match: 'null' do: [ block value: nil ] ]
! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 16:18'!
parseList
	self expectChar: $[.
	(self matchChar: $]) 
		ifTrue: [ ^ self listClass new ].
	self listClass streamContents: [ :stream |
		[ readStream atEnd ] 
			whileFalse: [
				stream nextPut: self parseValue.
				(self matchChar: $]) 
					ifTrue: [ ^ stream contents ].
				self expectChar: $, ] ].
	self error: 'end of list expected'

		! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 21:58'!
parseMap
	| map |
	map := self mapClass new.
	self expectChar: ${.
	(self matchChar: $})
		ifTrue: [ ^ map ].
	[ readStream atEnd ] whileFalse: [ | name value |
		name := self parsePropertyName.
		self expectChar: $:.
		value := self parseValue.
		map at: name put: value.
		(self matchChar: $}) 
			ifTrue: [ ^ map ].
		self expectChar: $, ].
	self error: 'end of map expected'! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 21:50'!
parseNumber
	| negated number |
	(negated := readStream peek = $-)
		ifTrue: [ readStream next ].
	number := self parseNumberInteger.
	(readStream peek = $.) 
		ifTrue: [
			readStream next. 
			number := number + self parseNumberFraction ].
	(readStream peek = $e or: [ readStream peek = $E ]) 
		ifTrue: [
			readStream next.
			number := number * self parseNumberExponent ].
	negated 
		ifTrue: [ number := number negated ].
	self consumeWhitespace.
	^ number! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 13:59'!
parseNumberExponent
	| number negated |
	number := 0.
	negated := readStream peek = $-.
	(negated or: [ readStream peek = $+ ]) 
		ifTrue: [ readStream next ].
	[ readStream atEnd not and: [ readStream peek isDigit ] ] 
		whileTrue: [ 
			number := 10 * number + readStream next digitValue ].
	negated 
		ifTrue: [ number := number negated ].
	^ 10 raisedTo: number! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 14:00'!
parseNumberFraction
	| number power |
	number := 0.
	power := 1.0.
	[ readStream atEnd not and: [ readStream peek isDigit ] ] 
		whileTrue: [
			number := 10 * number + readStream next digitValue.
			power := power * 10.0 ].
	^ number / power! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 14:00'!
parseNumberInteger
	| number |
	number := 0.
	[ readStream atEnd not and: [ readStream peek isDigit ] ] 
		whileTrue: [ 
			number := 10 * number + readStream next digitValue ].
	^ number! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 21:58'!
parsePropertyName
	| name |
	name := self parseValue.
	^ name isString
		ifTrue: [ name ]
		ifFalse: [ self error: 'property name must be string' ]! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 13:58'!
parseString
	| result |
	self expectChar: $".
	result := self stringStreamContents: [ :stream |
		[ readStream atEnd or: [ readStream peek = $"] ] 
			whileFalse: [ 
				stream nextPut: self parseCharacter ] ].
	self expectChar: $".
	^ result! !

!NeoJSONReader methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 6/2/2012 13:55'!
parseValue
	| char |
	readStream atEnd ifFalse: [ 
		char := readStream peek.
		char = ${
			ifTrue: [ ^ self parseMap ].
		char = $[
			ifTrue: [ ^ self parseList ].
		char = $"
			ifTrue: [ ^ self parseString ].
		(char = $- or: [ char isDigit ])
			ifTrue: [ ^ self parseNumber ].
		self parseConstantDo: [ :value | ^ value ] ].
	self error: 'invalid input'! !

!NeoJSONReader methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/2/2012 13:53'!
stringStreamContents: block
	stringStream ifNil: [ 
		stringStream := (String new: 32) writeStream ].
	stringStream reset.
	block value: stringStream.
	^ stringStream contents! !

!NeoJSONReader methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/2/2012 21:44'!
unescapeChar: char
	char = $b
		ifTrue: [ ^ Character backspace ].
	char = $f
		ifTrue: [ ^ Character newPage ].
	char = $n
		ifTrue: [ ^ Character lf ].
	char = $r
		ifTrue: [ ^ Character cr ].
	char = $t
		ifTrue: [ ^ Character tab ].
	self error: 'Unknown escape char: ', (String with: char)! !
